ble_address_type = {
    'gap_address_type_public': 0,
    'gap_address_type_random': 1
}
gap_discoverable_mode = {
    'non_discoverable': 0x00,
    'limited_discoverable': 0x01,
    'general_discoverable': 0x02,
    'broadcast': 0x03,
    'user_data': 0x04,
    'enhanced_broadcasting': 0x80
}
gap_connectable_mode = {
    'non_connectable': 0x00,
    'directed_connectable': 0x01,
    'undirected_connectable': 0x02,
    'scannable_non_connectable': 0x03,
}
gap_discover_mode = {
    'limited': 0x00,
    'generic': 0x01,
    'observation': 0x02,
}
bonding = {  # create bonding if devices not already bonded
    'do_not_create_bonding': 0x00,
    'create_bonding': 0x01,
}
bondable = {
    'no': 0x00,
    'yes': 0x01,
}
connection_status_flag = {
    'connected': 0x01,
    'encrypted': 0x02,
    'completed': 0x04,
    'parameters_change': 0x08,
}
scan_response_packet_type = {
    0x00: 'connectable_advertisement_packet',
    0x02: 'non-connectable_advertisement_packet',
    0x04: 'scan_response_packet',
    0x06: 'discoverable_advertisement_packet',
}
scan_response_data_type = {
    0x01: 'flags',
    0x02: 'incomplete_list_16-bit_service_class_uuids',
    0x03: 'complete_list_16-bit_service_class_uuids',
    0x04: 'incomplete_list_32-bit_service_class_uuids',
    0x05: 'complete_list_32-bit_service_class_uuids',
    0x06: 'incomplete_list_128-bit_service_class_uuids',
    0x07: 'complete_list_128-bit_service_class_uuids',
    0x08: 'shortened_local_name',
    0x09: 'complete_local_name',
    0x0A: 'tx_power_level',
    0x0D: 'class_of_device',
    0x0E: 'simple_pairing_hash_c/c-192',
    0x0F: 'simple_pairing_randomizer_r/r-192',
    0x10: 'device_id/security_manager_tk_value',
    0x11: 'security_manager_out_of_band_flags',
    0x12: 'slave_connection_interval_range',
    0x14: 'list_of_16-bit_service_solicitation_uuids',
    0x1F: 'list_of_32-bit_service_solicitation_uuids',
    0x15: 'list_of_128-bit_service_solicitation_uuids',
    0x16: 'service_data/service_data-16-bit_uuid',
    0x20: 'service_data-32-bit_uuid',
    0x21: 'service_data-128-bit_uuid',
    0x22: 'LE_secure_connections_confirmation_value',
    0x23: 'LE_secure_connections_random_value',
    0x17: 'public_target_address',
    0x18: 'random_target_address',
    0x19: 'appearance',
    0x1A: 'advertising_interval',
    0x1B: 'LE_bluetooth_device_address',
    0x1C: 'LE_role',
    0x1D: 'simple_pairing_hash_c-256',
    0x1E: 'simple_pairing_randomizer_r-256',
    0x3D: '3D_information_data',
    0xFF: 'manufacturer_specific_data',
}

# GATT
gatt_service_uuid = {
    'generic_access_profile': bytearray([0x18, 0x00]),
    'generic_attribute_profile': bytearray([0x18, 0x01]),
}
gatt_attribute_type_uuid = {
    'primary_service': bytearray([0x28, 0x00]),
    'secondary_service': bytearray([0x28, 0x01]),
    'include': bytearray([0x28, 0x02]),
    'characteristic': bytearray([0x28, 0x03]),
}
gatt_characteristic_descriptor_uuid = {
    'characteristic_extended_properties': bytearray([0x29, 0x00]),
    'characteristic_user_description': bytearray([0x29, 0x01]),
    'client_characteristic_configuration': bytearray([0x29, 0x02]),
    'server_characteristic_configuration': bytearray([0x29, 0x03]),
    'characteristic_format': bytearray([0x29, 0x04]),
    'characteristic_aggregate_format': bytearray([0x29, 0x05]),
}
gatt_characteristic_type_uuid = {
    'aerobic_heart_rate_lower_limit': bytearray([0x2A, 0x7E]),
    'aerobic_heart_rate_upper_limit': bytearray([0x2A, 0x84]),
    'aerobic_threshold': bytearray([0x2A, 0x7F]),
    'age': bytearray([0x2A, 0x80]),
    'aggregate': bytearray([0x2A, 0x5A]),
    'alert_category_id': bytearray([0x2A, 0x43]),
    'alert_category_id_bit_mask': bytearray([0x2A, 0x42]),
    'alert_level': bytearray([0x2A, 0x06]),
    'alert_notification_control_point': bytearray([0x2A, 0x44]),
    'alert_status': bytearray([0x2A, 0x3F]),
    'altitude': bytearray([0x2A, 0xB3]),
    'anaerobic_heart_rate_lower_limit': bytearray([0x2A, 0x81]),
    'anaerobic_heart_rate_upper_limit': bytearray([0x2A, 0x82]),
    'anaerobic_threshold': bytearray([0x2A, 0x83]),
    'analog': bytearray([0x2A, 0x58]),
    'apparent_wind_direction': bytearray([0x2A, 0x73]),
    'apparent_wind_speed': bytearray([0x2A, 0x72]),
    'appearance': bytearray([0x2A, 0x01]),
    'barometric_pressure_trend': bytearray([0x2A, 0xA3]),
    'battery_level': bytearray([0x2A, 0x19]),
    'blood_pressure_feature': bytearray([0x2A, 0x49]),
    'blood_pressure_measurement': bytearray([0x2A, 0x35]),
    'body_composition_feature': bytearray([0x2A, 0x9B]),
    'body_composition_measurement': bytearray([0x2A, 0x9C]),
    'body_sensor_location': bytearray([0x2A, 0x38]),
    'bond_management_control_point': bytearray([0x2A, 0xA4]),
    'bond_management_feature': bytearray([0x2A, 0xA5]),
    'boot_keyboard_input_report': bytearray([0x2A, 0x22]),
    'boot_keyboard_output_report': bytearray([0x2A, 0x32]),
    'boot_mouse_input_report': bytearray([0x2A, 0x33]),
    'central_address_resolution_support': bytearray([0x2A, 0xA6]),
    'cgm_feature': bytearray([0x2A, 0xA8]),
    'cgm_measurement': bytearray([0x2A, 0xA7]),
    'cgm_session_run_time': bytearray([0x2A, 0xAB]),
    'cgm_session_start_time': bytearray([0x2A, 0xAA]),
    'cgm_specific_ops_control_point': bytearray([0x2A, 0xAC]),
    'cgm_status': bytearray([0x2A, 0xA9]),
    'csc_feature': bytearray([0x2A, 0x5C]),
    'csc_measurement': bytearray([0x2A, 0x5B]),
    'current_time': bytearray([0x2A, 0x2B]),
    'cycling_power_control_point': bytearray([0x2A, 0x66]),
    'cycling_power_feature': bytearray([0x2A, 0x65]),
    'cycling_power_measurement': bytearray([0x2A, 0x63]),
    'cycling_power_vector': bytearray([0x2A, 0x64]),
    'database_change_increment': bytearray([0x2A, 0x99]),
    'date_of_birth': bytearray([0x2A, 0x85]),
    'date_of_threshold_assessment': bytearray([0x2A, 0x86]),
    'date_time': bytearray([0x2A, 0x08]),
    'day_date_time': bytearray([0x2A, 0x0A]),
    'day_of_week': bytearray([0x2A, 0x09]),
    'descriptor_value_changed': bytearray([0x2A, 0x7D]),
    'device_name': bytearray([0x2A, 0x00]),
    'dew_point': bytearray([0x2A, 0x7B]),
    'digital': bytearray([0x2A, 0x56]),
    'dst_offset': bytearray([0x2A, 0x0D]),
    'elevation': bytearray([0x2A, 0x6C]),
    'email_address': bytearray([0x2A, 0x87]),
    'exact_time_256': bytearray([0x2A, 0x0C]),
    'fat_burn_heart_rate_lower_limit': bytearray([0x2A, 0x88]),
    'fat_burn_heart_rate_upper_limit': bytearray([0x2A, 0x89]),
    'firmware_revision_string': bytearray([0x2A, 0x26]),
    'first_name': bytearray([0x2A, 0x8A]),
    'five_zone_heart_rate_limits': bytearray([0x2A, 0x8B]),
    'floor_number': bytearray([0x2A, 0xB2]),
    'gender': bytearray([0x2A, 0x8C]),
    'glucose_feature': bytearray([0x2A, 0x51]),
    'glucose_measurement': bytearray([0x2A, 0x18]),
    'glucose_measurement_context': bytearray([0x2A, 0x34]),
    'gust_factor': bytearray([0x2A, 0x74]),
    'hardware_revision_string': bytearray([0x2A, 0x27]),
    'heart_rate_control_point': bytearray([0x2A, 0x39]),
    'heart_rate_max': bytearray([0x2A, 0x8D]),
    'heart_rate_measurement': bytearray([0x2A, 0x37]),
    'heat_index': bytearray([0x2A, 0x7A]),
    'height': bytearray([0x2A, 0x8E]),
    'hid_control_point': bytearray([0x2A, 0x4C]),
    'hid_information': bytearray([0x2A, 0x4A]),
    'hip_circumference': bytearray([0x2A, 0x8F]),
    'http_control_point': bytearray([0x2A, 0xBA]),
    'http_entity_body': bytearray([0x2A, 0xB9]),
    'http_headers': bytearray([0x2A, 0xB7]),
    'http_status_code': bytearray([0x2A, 0xB8]),
    'https_security': bytearray([0x2A, 0xBB]),
    'humidity': bytearray([0x2A, 0x6F]),
    'ieee_11073-20601_regulatory_certification_data_list': bytearray([0x2A, 0x2A]),  # noqa
    'indoor_positioning_configuration': bytearray([0x2A, 0xAD]),
    'intermediate_cuff_pressure': bytearray([0x2A, 0x36]),
    'intermediate_temperature': bytearray([0x2A, 0x1E]),
    'irradiance': bytearray([0x2A, 0x77]),
    'language': bytearray([0x2A, 0xA2]),
    'last_name': bytearray([0x2A, 0x90]),
    'latitude': bytearray([0x2A, 0xAE]),
    'ln_control_point': bytearray([0x2A, 0x6B]),
    'ln_feature': bytearray([0x2A, 0x6A]),
    'local_east_coordinate': bytearray([0x2A, 0xB1]),
    'local_north_coordinate': bytearray([0x2A, 0xB0]),
    'local_time_information': bytearray([0x2A, 0x0F]),
    'location_and_speed': bytearray([0x2A, 0x67]),
    'location_name': bytearray([0x2A, 0xB5]),
    'longitude': bytearray([0x2A, 0xAF]),
    'magnetic_declination': bytearray([0x2A, 0x2C]),
    'magnetic_flux_density_2D': bytearray([0x2A, 0xA0]),
    'magnetic_flux_density_3D': bytearray([0x2A, 0xA1]),
    'manufacturer_name_string': bytearray([0x2A, 0x29]),
    'maximum_recommended_heart_rate': bytearray([0x2A, 0x91]),
    'measurement_interval': bytearray([0x2A, 0x21]),
    'model_number_string': bytearray([0x2A, 0x24]),
    'navigation': bytearray([0x2A, 0x68]),
    'new_alert': bytearray([0x2A, 0x46]),
    'object_action_control_point': bytearray([0x2A, 0xC5]),
    'object_changed': bytearray([0x2A, 0xC8]),
    'object_first_created': bytearray([0x2A, 0xC1]),
    'object_id': bytearray([0x2A, 0xC3]),
    'object_last_modified': bytearray([0x2A, 0xC2]),
    'object_list_control_point': bytearray([0x2A, 0xC6]),
    'object_list_filter': bytearray([0x2A, 0xC7]),
    'object_name': bytearray([0x2A, 0xBE]),
    'object_properties': bytearray([0x2A, 0xC4]),
    'object_size': bytearray([0x2A, 0xC0]),
    'object_type': bytearray([0x2A, 0xBF]),
    'ots_feature': bytearray([0x2A, 0xBD]),
    'peripheral_preferred_connection_parameters': bytearray([0x2A, 0x04]),
    'peripheral_privacy_flag': bytearray([0x2A, 0x02]),
    'plx_continuous_measurement': bytearray([0x2A, 0x5F]),
    'plx_features': bytearray([0x2A, 0x60]),
    'plx_spot_check_measurement': bytearray([0x2A, 0x5E]),
    'pnp_id': bytearray([0x2A, 0x50]),
    'pollen_concentration': bytearray([0x2A, 0x75]),
    'position_quality': bytearray([0x2A, 0x69]),
    'pressure': bytearray([0x2A, 0x6D]),
    'protocol_mode': bytearray([0x2A, 0x4E]),
    'rainfall': bytearray([0x2A, 0x78]),
    'reconnection_address': bytearray([0x2A, 0x03]),
    'record_access_control_point': bytearray([0x2A, 0x52]),
    'reference_time_information': bytearray([0x2A, 0x14]),
    'report': bytearray([0x2A, 0x4D]),
    'report_map': bytearray([0x2A, 0x4B]),
    'resting_heart_rate': bytearray([0x2A, 0x92]),
    'ringer_control_point': bytearray([0x2A, 0x40]),
    'ringer_setting': bytearray([0x2A, 0x41]),
    'rsc_feature': bytearray([0x2A, 0x54]),
    'rsc_measurement': bytearray([0x2A, 0x53]),
    'sc_control_point': bytearray([0x2A, 0x55]),
    'scan_interval_window': bytearray([0x2A, 0x4F]),
    'scan_refresh': bytearray([0x2A, 0x31]),
    'sensor_location': bytearray([0x2A, 0x5D]),
    'serial_number_string': bytearray([0x2A, 0x25]),
    'service_changed': bytearray([0x2A, 0x05]),
    'software_revision_string': bytearray([0x2A, 0x28]),
    'sport_type_for_aerobic_and_anaerobic_thresholds': bytearray([0x2A, 0x93]),
    'supported_new_alert_category': bytearray([0x2A, 0x47]),
    'supported_unread_alert_category': bytearray([0x2A, 0x48]),
    'system_id': bytearray([0x2A, 0x23]),
    'tds_control_point': bytearray([0x2A, 0xBC]),
    'temperature': bytearray([0x2A, 0x6E]),
    'temperature_measurement': bytearray([0x2A, 0x1C]),
    'temperature_type': bytearray([0x2A, 0x1D]),
    'three_zone_heart_rate_limits': bytearray([0x2A, 0x94]),
    'time_accuracy': bytearray([0x2A, 0x12]),
    'time_source': bytearray([0x2A, 0x13]),
    'time_update_control_point': bytearray([0x2A, 0x16]),
    'time_update_state': bytearray([0x2A, 0x17]),
    'time_with_dst': bytearray([0x2A, 0x11]),
    'time_zone': bytearray([0x2A, 0x0E]),
    'true_wind_direction': bytearray([0x2A, 0x71]),
    'true_wind_speed': bytearray([0x2A, 0x70]),
    'two_zone_heart_rate_limit': bytearray([0x2A, 0x95]),
    'tx_power_level': bytearray([0x2A, 0x07]),
    'uncertainty': bytearray([0x2A, 0xB4]),
    'unread_alert_status': bytearray([0x2A, 0x45]),
    'uri': bytearray([0x2A, 0xB6]),
    'user_control_point': bytearray([0x2A, 0x9F]),
    'user_index': bytearray([0x2A, 0x9A]),
    'uv_index': bytearray([0x2A, 0x76]),
    'vo2_max': bytearray([0x2A, 0x96]),
    'waist_circumference': bytearray([0x2A, 0x97]),
    'weight': bytearray([0x2A, 0x98]),
    'weight_measurement': bytearray([0x2A, 0x9D]),
    'weight_scale_feature': bytearray([0x2A, 0x9E]),
    'wind_chill': bytearray([0x2A, 0x79]),
}



import logging
import time

from pygatt import BLEDevice, exceptions
from . import constants
from .bgapi import BGAPIError
from .error_codes import ErrorCode
from .packets import BGAPICommandPacketBuilder as CommandBuilder
from .bglib import EventPacketType, ResponsePacketType

log = logging.getLogger(__name__)


def connection_required(func):
    """Raise an exception if the device is not connected before calling the
    actual function.
    """
    def wrapper(self, *args, **kwargs):
        if self._handle is None:
            raise exceptions.NotConnectedError()
        return func(self, *args, **kwargs)
    return wrapper


class BGAPIBLEDevice(BLEDevice):
    def __init__(self, address, handle, backend):
        super(BGAPIBLEDevice, self).__init__(address)
        self._handle = handle
        self._backend = backend

    @connection_required
    def bond(self, permanent=False):
        """
        Create a bond and encrypted connection with the device.
        """

        # Set to bondable mode so bonds are store permanently
        if permanent:
            self._backend.set_bondable(True)
        log.debug("Bonding to %s", self._address)
        self._backend.send_command(
            CommandBuilder.sm_encrypt_start(
                self._handle, constants.bonding['create_bonding']))
        self._backend.expect(ResponsePacketType.sm_encrypt_start)

        packet_type, response = self._backend.expect_any(
            [EventPacketType.connection_status,
             EventPacketType.sm_bonding_fail])
        if packet_type == EventPacketType.sm_bonding_fail:
            raise BGAPIError("Bonding failed")
        log.info("Bonded to %s", self._address)

    @connection_required
    def get_rssi(self):
        """
        Get the receiver signal strength indicator (RSSI) value from the device.

        Returns the RSSI as in integer in dBm.
        """
        # The BGAPI has some strange behavior where it will return 25 for
        # the RSSI value sometimes... Try a maximum of 3 times.
        for i in range(0, 3):
            self._backend.send_command(
                CommandBuilder.connection_get_rssi(self._handle))
            _, response = self._backend.expect(
                ResponsePacketType.connection_get_rssi)
            rssi = response['rssi']
            if rssi != 25:
                return rssi
            time.sleep(0.1)
        raise BGAPIError("get rssi failed")

    @connection_required
    def char_read(self, uuid, timeout=None):
        return self.char_read_handle(self.get_handle(uuid), timeout=timeout)

    @connection_required
    def char_read_handle(self, handle, timeout=None):
        log.info("Reading characteristic at handle %d", handle)
        self._backend.send_command(
            CommandBuilder.attclient_read_by_handle(
                self._handle, handle))

        self._backend.expect(ResponsePacketType.attclient_read_by_handle)
        success = False
        while not success:
            matched_packet_type, response = self._backend.expect_any(
                [EventPacketType.attclient_attribute_value,
                 EventPacketType.attclient_procedure_completed],
                timeout=timeout)
            # TODO why not just expect *only* the attribute value response,
            # then it would time out and raise an exception if allwe got was
            # the 'procedure completed' response?
            if matched_packet_type != EventPacketType.attclient_attribute_value:
                raise BGAPIError("Unable to read characteristic")
            if response['atthandle'] == handle:
                # Otherwise we received a response from a wrong handle (e.g.
                # from a notification) so we keep trying to wait for the
                # correct one
                success = True
        return bytearray(response['value'])

    @connection_required
    def char_read_long(self, uuid, timeout=None):
        return self.char_read_long_handle(self.get_handle(uuid),
                                          timeout=timeout)

    @connection_required
    def char_read_long_handle(self, handle, timeout=None):
        log.info("Reading long characteristic at handle %d", handle)
        self._backend.send_command(
            CommandBuilder.attclient_read_long(
                self._handle, handle))

        self._backend.expect(ResponsePacketType.attclient_read_long)
        success = False
        response = b""
        while not success:
            matched_packet_type, chunk = self._backend.expect_any(
                [EventPacketType.attclient_attribute_value,
                 EventPacketType.attclient_procedure_completed],
                timeout=timeout)

            if (matched_packet_type ==
                    EventPacketType.attclient_attribute_value):
                if chunk['atthandle'] == handle:
                    # Concatenate the data
                    response += chunk["value"]
            elif (matched_packet_type ==
                    EventPacketType.attclient_procedure_completed):
                if chunk['chrhandle'] == handle:
                    success = True
        return bytearray(response)

    @connection_required
    def char_write_handle(self, char_handle, value, wait_for_response=True):
        while True:
            value_list = [b for b in value]
            # An "attribute write" is always acknowledged by the remote host.
            if wait_for_response:
                self._backend.send_command(
                    CommandBuilder.attclient_attribute_write(
                        self._handle, char_handle, value_list))
                self._backend.expect(
                    ResponsePacketType.attclient_attribute_write)
                packet_type, response = self._backend.expect(
                    EventPacketType.attclient_procedure_completed,
                    # According to the BLE spec, the device has 30 seconds to
                    # repsonse to the attribute write.
                    timeout=30)

            # A "command" write is unacknowledged - don't wait for a response.
            else:
                self._backend.send_command(
                    CommandBuilder.attclient_write_command(
                        self._handle, char_handle, value_list))
                packet_type, response = self._backend.expect(
                    ResponsePacketType.attclient_write_command)

            if (response['result'] !=
                    ErrorCode.insufficient_authentication.value):
                # Continue to retry until we are bonded
                break

    # ASC - adapted from
    # https://raw.githubusercontent.com/mjbrown/bgapi/master/bgapi/module.py
    # - reliable_write_by_handle
    @connection_required
    def char_write_long_handle(self,
                               char_handle,
                               value,
                               wait_for_response=False):

        maxv = 18

        for i in range(int(((len(value)-1) / maxv)+1)):

            chunk = value[maxv*i:min(maxv*(i+1), len(value))]
            value_list = [b for b in chunk]
            self._backend.send_command(
                CommandBuilder.attclient_prepare_write(
                    self._handle, char_handle, maxv*i, value_list))

            packet_type, response = self._backend.expect(
                ResponsePacketType.attclient_prepare_write)

            packet_type, response = self._backend.expect(
                EventPacketType.attclient_procedure_completed)
            time.sleep(0.1)

        time.sleep(0.1)
        self._backend.send_command(
            CommandBuilder.attclient_execute_write(
                self._handle, 1))  # 1 = commit, 0 = cancel
        self._backend.expect(ResponsePacketType.attclient_execute_write)
        packet_type, response = self._backend.expect(
            EventPacketType.attclient_procedure_completed)
        time.sleep(0.1)

    @connection_required
    def disconnect(self):
        log.debug("Disconnecting from %s", self._address)
        self._backend.send_command(
            CommandBuilder.connection_disconnect(self._handle))

        self._backend.expect(ResponsePacketType.connection_disconnect)
        log.info("Disconnected from %s", self._address)
        self._handle = None

    @connection_required
    def discover_characteristics(self):
        self._characteristics = self._backend.discover_characteristics(
            self._handle)
        return self._characteristics



from __future__ import print_function


# for Python 2/3 compatibility
try:
    import queue
except ImportError:
    import Queue as queue

import logging
import serial
import time
import threading
from binascii import hexlify, unhexlify
from uuid import UUID
from enum import Enum
from collections import defaultdict

from pygatt.exceptions import NotConnectedError
from pygatt.backends import BLEBackend, Characteristic, BLEAddressType
from pygatt.util import uuid16_to_uuid

from . import bglib, constants
from .exceptions import BGAPIError, ExpectedResponseTimeout
from .device import BGAPIBLEDevice
from .bglib import EventPacketType, ResponsePacketType
from .packets import BGAPICommandPacketBuilder as CommandBuilder
from .error_codes import get_return_message
from .util import find_usb_serial_devices

try:
    import termios
except ImportError:
    # Running in Windows (not Linux/OS X/Cygwin)
    serial_exceptions = (RuntimeError, ValueError,
                         BGAPIError, serial.serialutil.SerialException)
else:
    serial_exceptions = (termios.error, IOError, OSError, TypeError,
                         BGAPIError, serial.serialutil.SerialException)


log = logging.getLogger(__name__)

BLED112_VENDOR_ID = 0x2458
BLED112_PRODUCT_ID = 0x0001
MAX_CONNECTION_ATTEMPTS = 10


UUIDType = Enum('UUIDType', ['custom', 'service', 'attribute',
                             'descriptor', 'characteristic',
                             'nonstandard'])


def _timed_out(start_time, timeout):
    return time.time() - start_time > timeout


def bgapi_address_to_hex(address):
    address = hexlify(bytearray(
        list(reversed(address)))).upper().decode('ascii')
    return ':'.join(''.join(pair) for pair in zip(*[iter(address)] * 2))


class AdvertisingAndScanInfo(object):
    """
    Holds the advertising and scan response packet data from a device at a given
    address.
    """
    def __init__(self):
        self.name = ""
        self.address = ""
        self.rssi = None
        self.packet_data = {
            # scan_response_packet_type[xxx]: data_dictionary,
        }


class BGAPIBackend(BLEBackend):
    """
    A BLE backend for a BGAPI compatible USB adapter.
    """
    def __init__(self, serial_port=None, receive_queue_timeout=0.1):
        """
        Initialize the backend, but don't start the USB connection yet. Must
        call .start().

        serial_port -- The name of the serial port for the BGAPI-compatible
            USB interface. If not provided, will attempt to auto-detect.
        """
        self._lib = bglib.BGLib()
        self._serial_port = serial_port
        self._receive_queue_timeout = receive_queue_timeout

        self._ser = None
        self._receiver = None
        self._running = None
        self._lock = threading.Lock()
        self._evt = threading.Event()
        self._scan_cb = None

        # buffer for packets received
        self._receiver_queue = queue.Queue()

        # State
        self._num_bonds = 0  # number of bonds stored on the adapter
        self._stored_bonds = []  # bond handles stored on the adapter
        self._devices_discovered = {
            # 'address': AdvertisingAndScanInfo,
            # Note: address formatted like "01:23:45:67:89:AB"
        }
        self._characteristics = defaultdict(dict)
        self._connections = {}

        self._current_characteristic = None  # used in char/descriptor discovery
        self._packet_handlers = {
            ResponsePacketType.sm_get_bonds: self._ble_rsp_sm_get_bonds,
            ResponsePacketType.system_address_get: (
                self._ble_rsp_system_address_get),
            EventPacketType.attclient_attribute_value: (
                self._ble_evt_attclient_attribute_value),
            EventPacketType.attclient_find_information_found: (
                self._ble_evt_attclient_find_information_found),
            EventPacketType.connection_status: self._ble_evt_connection_status,
            EventPacketType.connection_disconnected: (
                self._ble_evt_connection_disconnected),
            EventPacketType.gap_scan_response: self._ble_evt_gap_scan_response,
            EventPacketType.sm_bond_status: self._ble_evt_sm_bond_status,
        }

        log.info("Initialized new BGAPI backend")

    def _detect_device_port(self):
        if self._serial_port:
            try:
                serial.Serial(port=self._serial_port)
                return self._serial_port
            except serial.serialutil.SerialException:
                raise BGAPIError(
                    "Unable to detect BLED112 serial port: {}.".format(
                        self._serial_port))
        else:
            log.info("Auto-detecting serial port for BLED112")
            detected_devices = find_usb_serial_devices(
                vendor_id=BLED112_VENDOR_ID,
                product_id=BLED112_PRODUCT_ID)
            if len(detected_devices) == 0:
                raise BGAPIError("Unable to auto-detect BLED112 serial port")

            log.info("Found BLED112 on serial port %s",
                     detected_devices[0].port_name)
            return detected_devices[0].port_name

    def _open_serial_port(self,
                          max_connection_attempts=MAX_CONNECTION_ATTEMPTS):
        """
        Open a connection to the named serial port, or auto-detect the first
        port matching the BLED device. This will wait until data can actually be
        read from the connection, so it will not return until the device is
        fully booted.

        max_connection_attempts -- Max number of times to retry
            detecting and connecting to a device.

        Raises a NotConnectedError if the device cannot connect after 10
        attempts, with a short pause in between each attempt.
        """
        for attempt in range(max_connection_attempts):
            log.debug("Opening connection to serial port (attempt %d)",
                      attempt + 1)
            try:
                serial_port = self._serial_port or self._detect_device_port()
                self._ser = None

                self._ser = serial.Serial(serial_port, baudrate=115200,
                                          timeout=0.25)
                # Wait until we can actually read from the device
                self._ser.read()
                break
            except serial_exceptions:
                log.debug("Failed to open serial port", exc_info=True)
                if self._ser:
                    self._ser.close()
                elif attempt == (max_connection_attempts - 1):
                    raise NotConnectedError(
                        "No BGAPI compatible device detected")
                self._ser = None
                time.sleep(0.25)
        else:
            raise NotConnectedError("Unable to reconnect with USB "
                                    "device after rebooting")

    def start(self, reset=True):
        """
        Connect to the USB adapter, reset its state and start a background
        receiver thread.

        reset - Reset the device if True.
        """
        if self._running and self._running.is_set():
            self.stop()

        # Fail immediately if no device is attached
        self._detect_device_port()

        # Blow everything away and start anew.
        # Only way to be sure is to burn it down and start again.
        # (Aka reset remote state machine)

        # The zero param just means we want to do a normal restart instead of
        # starting a firmware update restart.
        if reset:
            log.info(
                "Resetting and reconnecting to device for a clean environment")
            self._open_serial_port()
            self.send_command(CommandBuilder.system_reset(0))
            self._ser.close()

            # Wait before re-opening the port - required on at least Windows,
            # possibly OS X.
            time.sleep(0.5)

        self._open_serial_port()
        self._receiver = threading.Thread(target=self._receive)
        self._receiver.daemon = True

        self._running = threading.Event()
        self._running.set()
        self._receiver.start()

        self.disable_advertising()
        self.set_bondable(False)

        # Stop any ongoing procedure
        log.debug("Stopping any outstanding GAP procedure")
        self.send_command(CommandBuilder.gap_end_procedure())
        try:
            self.expect(ResponsePacketType.gap_end_procedure)
        except BGAPIError:
            # Ignore any errors if there was no GAP procedure running
            pass

    def get_mac(self):
        self.send_command(CommandBuilder.system_address_get())
        self.expect(ResponsePacketType.system_address_get)

    def stop(self):
        for device in self._connections.values():
            try:
                device.disconnect()
            except NotConnectedError:
                pass
        if self._running:
            if self._running.is_set():
                log.info('Stopping')
            self._running.clear()

        if self._receiver:
            self._receiver.join()
        self._receiver = None

        if self._ser:
            self._ser.close()
            self._ser = None

    def set_bondable(self, bondable):
        self.send_command(
            CommandBuilder.sm_set_bondable_mode(
                constants.bondable['yes' if bondable else 'no']))
        self.expect(ResponsePacketType.sm_set_bondable_mode)

    def disable_advertising(self):
        log.info("Disabling advertising")
        self.send_command(
            CommandBuilder.gap_set_mode(
                constants.gap_discoverable_mode['non_discoverable'],
                constants.gap_connectable_mode['non_connectable']))
        self.expect(ResponsePacketType.gap_set_mode)

    def send_command(self, *args, **kwargs):
        with self._lock:
            if self._ser is None:
                log.warn("Unexpectedly not connected to USB device")
                raise NotConnectedError()
            return self._lib.send_command(self._ser, *args, **kwargs)

    def clear_bond(self, address=None):
        """
        Delete the bonds stored on the adapter.

        address - the address of the device to unbond. If not provided, will
            erase all bonds.

        Note: this does not delete the corresponding bond stored on the remote
              device.
        """
        # Find bonds
        log.info("Fetching existing bonds for devices")
        self._stored_bonds = []
        self.send_command(CommandBuilder.sm_get_bonds())

        try:
            self.expect(ResponsePacketType.sm_get_bonds)
        except NotConnectedError:
            pass

        if self._num_bonds == 0:
            return

        while len(self._stored_bonds) < self._num_bonds:
            self.expect(EventPacketType.sm_bond_status)

        for b in reversed(self._stored_bonds):
            log.info("Deleting bond %s", b)

            self.send_command(CommandBuilder.sm_delete_bonding(b))
            self.expect(ResponsePacketType.sm_delete_bonding)

    def scan(self, timeout=10, scan_interval=75, scan_window=50, active=True,
             discover_mode=constants.gap_discover_mode['observation'],
             scan_cb=None, **kwargs):
        """
        Perform a scan to discover BLE devices.

        timeout -- the number of seconds this scan should last.
        scan_interval -- the number of milliseconds until scanning is restarted.
        scan_window -- the number of milliseconds the scanner will listen on one
                     frequency for advertisement packets.
        active -- True --> ask sender for scan response data. False --> don't.
        discover_mode -- one of the gap_discover_mode constants.
        scan_cb -- This callback function is called whenever a new BLE
                   advertising packet is received.
                   The function takes three parameters:
                       devices, addr, packet_type
                   If the function returns True, the scan is aborted
        """
        self._scan_cb = scan_cb
        parameters = 1 if active else 0
        # NOTE: the documentation seems to say that the times are in units of
        # 625us but the ranges it gives correspond to units of 1ms....
        self.send_command(
            CommandBuilder.gap_set_scan_parameters(
                scan_interval, scan_window, parameters
            ))

        self.expect(ResponsePacketType.gap_set_scan_parameters)

        log.info("Starting an %s scan", "active" if active else "passive")
        self.send_command(CommandBuilder.gap_discover(discover_mode))

        self.expect(ResponsePacketType.gap_discover)

        log.info("Pausing for maximum %ds to allow scan to complete", timeout)

        self._evt.set()
        start_time = time.time()

        while self._evt.is_set():
            try:
                self.expect(EventPacketType.gap_scan_response,
                            timeout=timeout)
            except ExpectedResponseTimeout:
                pass
            if _timed_out(start_time, timeout):
                break

        log.info("Stopping scan")
        self.send_command(CommandBuilder.gap_end_procedure())
        self.expect(ResponsePacketType.gap_end_procedure)

        devices = []
        for address, info in self._devices_discovered.items():
            devices.append({
                'address': address,
                'name': info.name,
                'rssi': info.rssi,
                'packet_data': info.packet_data
            })
        log.info("Discovered %d devices: %s", len(devices), devices)
        self._devices_discovered = {}
        return devices

    def _end_procedure(self):
        self.send_command(CommandBuilder.gap_end_procedure())
        self.expect(ResponsePacketType.gap_end_procedure)

    def connect(self, address, timeout=5,
                address_type=BLEAddressType.public,
                interval_min=60, interval_max=76, supervision_timeout=100,
                latency=0):
        """
        Connect directly to a device given the ble address then discovers and
        stores the characteristic and characteristic descriptor handles.

        Requires that the adapter is not connected to a device already.

        address -- a bytearray containing the device mac address.
        timeout -- number of seconds to wait before returning if not connected.
        address_type -- one of BLEAddressType's values, either public or random.

        Raises BGAPIError or NotConnectedError on failure.
        """

        address_bytes = bytearray(unhexlify(address.replace(":", "")))
        for device in self._connections.values():
            if device._address == bgapi_address_to_hex(address_bytes):
                return device

        log.info("Connecting to device at address %s (timeout %ds)",
                 address, timeout)
        self.set_bondable(False)

        if address_type == BLEAddressType.public:
            addr_type = constants.ble_address_type['gap_address_type_public']
        else:
            addr_type = constants.ble_address_type['gap_address_type_random']

        self.send_command(
            CommandBuilder.gap_connect_direct(
                address_bytes, addr_type, interval_min, interval_max,
                supervision_timeout, latency))

        try:
            self.expect(ResponsePacketType.gap_connect_direct)
            _, packet = self.expect(EventPacketType.connection_status,
                                    timeout=timeout)
            # TODO what do we do if the status isn't 'connected'? Retry?
            # Raise an exception? Should also check the address matches the
            # expected TODO i'm finding that when reconnecting to the same
            # MAC, we geta conneciotn status of "disconnected" but that is
            # picked up here as "connected", then we don't get anything
            # else.
            if self._connection_status_flag(
                    packet['flags'],
                    constants.connection_status_flag['connected']):
                device = BGAPIBLEDevice(
                    bgapi_address_to_hex(packet['address']),
                    packet['connection_handle'],
                    self)
                if self._connection_status_flag(
                        packet['flags'],
                        constants.connection_status_flag['encrypted']):
                    device.encrypted = True
                self._connections[packet['connection_handle']] = device
                log.info("Connected to %s", address)
                return device
        except ExpectedResponseTimeout:
            # If the connection doesn't occur because the device isn't there
            # then you should manually stop the command.
            #
            # If we never get the connection status it is likely that it
            # didn't occur because the device isn't there. If that is true
            # then we have to manually stop the command.
            self._end_procedure()
            exc = NotConnectedError()
            exc.__cause__ = None
            raise exc

    def discover_characteristics(self, connection_handle):
        att_handle_start = 0x0001  # first valid handle
        att_handle_end = 0xFFFF  # last valid handle
        log.info("Fetching characteristics for connection %d",
                 connection_handle)
        self.send_command(
            CommandBuilder.attclient_find_information(
                connection_handle, att_handle_start, att_handle_end))

        self.expect(ResponsePacketType.attclient_find_information)
        self.expect(EventPacketType.attclient_procedure_completed,
                    timeout=10)

        for char_uuid_str, char_obj in (
                self._characteristics[connection_handle].items()):
            log.info("Characteristic 0x%s is handle 0x%x",
                     char_uuid_str, char_obj.handle)
            for desc_uuid_str, desc_handle in (
                    char_obj.descriptors.items()):
                log.info("Characteristic descriptor 0x%s is handle 0x%x",
                         desc_uuid_str, desc_handle)
        return self._characteristics[connection_handle]

    @staticmethod
    def _connection_status_flag(flags, flag_to_find):
        """
        Is the given flag in the connection status flags?

        flags -- the 'flags' parameter returned by ble_evt_connection_status.
        flag_to_find -- the flag to look for in flags.

        Returns true if flag_to_find is in flags. Returns false otherwise.
        """
        return (flags & flag_to_find) == flag_to_find

    @staticmethod
    def _get_uuid_type(uuid):
        """
        Checks if the UUID is a custom 128-bit UUID or a GATT characteristic
        descriptor UUID.

        uuid -- the UUID as a bytearray.

        Return a UUIDType.
        """
        if len(uuid) == 16:  # 128-bit --> 16 byte
            return UUIDType.custom
        if uuid in constants.gatt_service_uuid.values():
            return UUIDType.service
        if uuid in constants.gatt_attribute_type_uuid.values():
            return UUIDType.attribute
        if uuid in constants.gatt_characteristic_descriptor_uuid.values():
            return UUIDType.descriptor
        if uuid in constants.gatt_characteristic_type_uuid.values():
            return UUIDType.characteristic

        log.warn("Unrecognized 4 byte UUID %s", hexlify(uuid))
        return UUIDType.nonstandard

    def _scan_rsp_data(self, data):
        """
        Parse scan response data.
        Note: the data will come in a format like the following:
        [data_length, data_type, data..., data_length, data_type, data...]

        data -- the args['data'] list from _ble_evt_scan_response.

        Returns a name and a dictionary containing the parsed data in pairs of
        field_name': value.
        """
        # Result stored here
        data_dict = {
            # 'name': value,
        }
        bytes_left_in_field = 0
        field_name = None
        field_value = []
        # Iterate over data bytes to put in field
        dev_name = ""
        for b in data:
            if bytes_left_in_field == 0:
                # New field
                bytes_left_in_field = b
                field_value = []
            else:
                field_value.append(b)
                bytes_left_in_field -= 1
                if bytes_left_in_field == 0:
                    # End of field
                    field_name = (
                        constants.scan_response_data_type[field_value[0]])
                    field_value = field_value[1:]
                    # Field type specific formats
                    if (field_name == 'complete_local_name' or
                            field_name == 'shortened_local_name'):
                        dev_name = bytearray(field_value).decode("utf-8")
                        data_dict[field_name] = dev_name
                    elif (field_name ==
                          'complete_list_128-bit_service_class_uuids'):
                        if len(field_value) % 16 == 0:  # 16 bytes
                            data_dict[field_name] = []
                            for i in range(0, int(len(field_value) / 16)):
                                service_uuid = (
                                    "0x%s" %
                                    bgapi_address_to_hex(
                                        field_value[i * 16:i * 16 + 16]))
                                data_dict[field_name].append(service_uuid)
                        else:
                            log.warning("Expected a service class UUID of 16\
                                        bytes. Instead received %d bytes",
                                        len(field_value))
                    else:
                        data_dict[field_name] = bytearray(field_value)
        return dev_name, data_dict

    def expect(self, expected, *args, **kargs):
        return self.expect_any([expected], *args, **kargs)

    def expect_any(self, expected_packet_choices, timeout=None,
                   assert_return_success=True):
        """
        Process packets until a packet of one of the expected types is found.

        expected_packet_choices -- a list of BGLib.PacketType.xxxxx. Upon
                                   processing a packet of a type contained in
                                   the list, this function will return.
        timeout -- maximum time in seconds to process packets.
        assert_return_success -- raise an exception if the return code from a
            matched message is non-zero.

        Raises an ExpectedResponseTimeout if one of the expected responses is
            not receiving withint the time limit.
        """
        timeout = timeout or 1
        log.debug("Expecting a response of one of %s within %fs",
                  expected_packet_choices, timeout or 0)

        start_time = None
        if timeout is not None:
            start_time = time.time()

        while True:
            packet = None
            try:
                packet = self._receiver_queue.get(
                    timeout=self._receive_queue_timeout)
            except queue.Empty:
                if timeout is not None:
                    if _timed_out(start_time, timeout):
                        exc = ExpectedResponseTimeout(
                            expected_packet_choices, timeout)
                        exc.__cause__ = None
                        raise exc
                    continue

            if packet is None:
                raise ExpectedResponseTimeout(expected_packet_choices, timeout)

            packet_type, response = self._lib.decode_packet(packet)
            return_code = response.get('result', 0)
            log.debug("Received a %s packet: %s",
                      packet_type, get_return_message(return_code))

            if packet_type in self._packet_handlers:
                self._packet_handlers[packet_type](response)

            if packet_type in expected_packet_choices:
                return packet_type, response

    def _receive(self):
        """
        Read bytes from serial and enqueue the packets if the packet is not a.
        Stops if the self._running event is not set.
        """
        log.info("Running receiver")
        while self._running.is_set():
            packet = self._lib.parse_byte(self._ser.read())
            if packet is not None:
                packet_type, args = self._lib.decode_packet(packet)
                if packet_type == EventPacketType.attclient_attribute_value:
                    device = self._connections[args['connection_handle']]
                    device.receive_notification(args['atthandle'],
                                                bytearray(args['value']))
                self._receiver_queue.put(packet)
        log.info("Stopping receiver")

    def _ble_evt_attclient_attribute_value(self, args):
        """
        Handles the event for values of characteristics.

        args -- dictionary containing the attribute handle ('atthandle'),
        attribute type ('type'), and attribute value ('value')
        """
        log.debug("attribute handle = %x", args['atthandle'])
        log.debug("attribute type = %x", args['type'])
        log.debug("attribute value = 0x%s", hexlify(bytearray(args['value'])))

    def _ble_evt_attclient_find_information_found(self, args):
        """
        Handles the event for characteritic discovery.

        Adds the characteristic to the dictionary of characteristics or adds
        the descriptor to the dictionary of descriptors in the current
        characteristic. These events will be occur in an order similar to the
        following:
        1) primary service uuid
        2) 0 or more descriptors
        3) characteristic uuid
        4) 0 or more descriptors
        5) repeat steps 3-4

        args -- dictionary containing the characteristic handle ('chrhandle'),
        and characteristic UUID ('uuid')
        """
        raw_uuid = bytearray(reversed(args['uuid']))

        # Convert 4-byte UUID shorthand to a full, 16-byte UUID
        uuid_type = self._get_uuid_type(raw_uuid)
        if uuid_type != UUIDType.custom:
            uuid = uuid16_to_uuid(int(
                bgapi_address_to_hex(args['uuid']).replace(':', ''), 16))
        else:
            uuid = UUID(bytes=bytes(raw_uuid))

        # TODO is there a way to get the characteristic from the packet instead
        # of having to track the "current" characteristic?
        if (uuid_type == UUIDType.descriptor and
                self._current_characteristic is not None):
            self._current_characteristic.add_descriptor(uuid, args['chrhandle'])
        elif (uuid_type == UUIDType.custom or
                uuid_type == UUIDType.nonstandard or
                uuid_type == UUIDType.characteristic):
            if uuid_type == UUIDType.custom:
                log.info("Found custom characteristic %s" % uuid)
            elif uuid_type == UUIDType.characteristic:
                log.info("Found approved characteristic %s" % uuid)
            elif uuid_type == UUIDType.nonstandard:
                log.info("Found nonstandard 4-byte characteristic %s" % uuid)
            new_char = Characteristic(uuid, args['chrhandle'])
            self._current_characteristic = new_char
            self._characteristics[
                args['connection_handle']][uuid] = new_char

    def _ble_evt_connection_disconnected(self, args):
        """
        Handles the event for the termination of a connection.
        """
        self._connections.pop(args['connection_handle'], None)

    def _ble_evt_connection_status(self, args):
        """
        Handles the event for reporting connection status.

        args -- dictionary containing the connection status flags ('flags'),
            device address ('address'), device address type ('address_type'),
            connection interval ('conn_interval'), connection timeout
            (timeout'), device latency ('latency'), device bond handle
            ('bonding')
        """
        connection_handle = args['connection_handle']
        if not self._connection_status_flag(
                args['flags'],
                constants.connection_status_flag['connected']):
            # Disconnected
            self._connections.pop(connection_handle, None)

        log.info("Connection status: handle=0x%x, flags=%s, address=0x%s, "
                 "connection interval=%fms, timeout=%d, "
                 "latency=%d intervals, bonding=0x%x",
                 connection_handle,
                 args['address'],
                 hexlify(bytearray(args['address'])),
                 args['conn_interval'] * 1.25,
                 args['timeout'] * 10,
                 args['latency'],
                 args['bonding'])

    def _ble_evt_gap_scan_response(self, args):
        """
        Handles the event for reporting the contents of an advertising or scan
        response packet.
        This event will occur during device discovery but not direct connection.

        args -- dictionary containing the RSSI value ('rssi'), packet type
                ('packet_type'), address of packet sender ('sender'), address
                type ('address_type'), existing bond handle ('bond'), and
                scan resonse data list ('data')
        """
        # Parse packet
        packet_type = constants.scan_response_packet_type[args['packet_type']]
        address = bgapi_address_to_hex(args['sender'])
        name, data_dict = self._scan_rsp_data(args['data'])

        # Store device information
        if address not in self._devices_discovered:
            self._devices_discovered[address] = AdvertisingAndScanInfo()
        dev = self._devices_discovered[address]
        if dev.name == "":
            dev.name = name
        if dev.address == "":
            dev.address = address
        if (packet_type not in dev.packet_data or
                len(dev.packet_data[packet_type]) < len(data_dict)):
            dev.packet_data[packet_type] = data_dict
        dev.rssi = args['rssi']
        log.debug("Received a scan response from %s with rssi=%d dBM "
                  "and data=%s", address, args['rssi'], data_dict)

        if self._scan_cb is not None:
            if self._scan_cb(self._devices_discovered, address, packet_type):
                self._evt.clear()

    def _ble_evt_sm_bond_status(self, args):
        """
        Handles the event for reporting a stored bond.

        Adds the stored bond to the list of bond handles.

        args -- dictionary containing the bond handle ('bond'), encryption key
                size used in the long-term key ('keysize'), was man in the
                middle used ('mitm'), keys stored for bonding ('keys')
        """
        # Add to list of stored bonds found or set flag
        self._stored_bonds.append(args['bond'])

    def _ble_rsp_sm_delete_bonding(self, args):
        """
        Handles the response for the deletion of a stored bond.

        args -- dictionary containing the return code ('result')
        """
        result = args['result']
        if result == 0:
            self._stored_bonds.pop()
        return result

    def _ble_rsp_sm_get_bonds(self, args):
        """
        Handles the response for the start of stored bond enumeration. Sets
        self._num_bonds to the number of stored bonds.

        args -- dictionary containing the number of stored bonds ('bonds'),
        """
        self._num_bonds = args['bonds']
        log.debug("num bonds = %d", args['bonds'])

    def _ble_rsp_system_address_get(self, args):
        """
        Handles the response for the system mac address. Stores the
        result as a member.

        args -- dictionary containing the mac address ('address'),
        """
        self.address = args['address']
        log.debug("Adapter address = {0}".format(args['address']))



from __future__ import print_function

from enum import Enum


class ErrorCode(Enum):
    insufficient_authentication = 0x0405


return_codes = {
    0: "Success",
    # BGAPI errors
    0x0180: "Invalid parameter",
    0x0181: "Device in wrong state",
    0x0182: "Out of memory",
    0x0183: "Feature not implemented",
    0x0184: "Command not recognized",
    0x0185: "Timeout",
    0x0186: "Not connected",
    0x0187: "Overflow/underflow",
    0x0188: "User attribute",
    0x0189: "Invalid license key",
    0x018A: "Command too long",
    0x018B: "Out of bounds",
    # Bluetooth errors
    0x0205: "Authentication failure",
    0x0206: "Pin or key missing",
    0x0207: "Memory capacity exceeded",
    0x0208: "Connection timeout",
    0x0209: "Connection limit exceeded",
    0x020C: "Command disallowed",
    0x0212: "Invalid command parameters",
    0x0213: "Remote user terminated connection",
    0x0216: "Connection terminated by local host",
    0x0222: "Link layer response timeout",
    0x0228: "Link layer instance passed",
    0x023A: "Controller busy",
    0x023B: "Unacceptable connection interval",
    0x023C: "Directed advertising timeout",
    0x023D: "MIC failure",
    0x023E: "Connection failed to be established",
    # Security manager protocol errors
    0x0301: "Passkey entry failed",
    0x0302: "OOB data is not available",
    0x0303: "Authentication requirements",
    0x0304: "Confirm value failed",
    0x0305: "Pairing not supported",
    0x0306: "Encryption key size",
    0x0307: "Command not supported",
    0x0308: "Unspecified reason",
    0x0309: "Repeated attempts",
    0x030A: "Invalid parameters",
    # Attribute protocol errors
    0x0401: "Invalid handle",
    0x0402: "Read not permitted",
    0x0403: "Write not permitted",
    0x0404: "Invalid PDU",
    ErrorCode.insufficient_authentication.value: "Insufficient authentication",
    0x0406: "Request not supported",
    0x0407: "Invalid offset",
    0x0408: "Insufficient authorization",
    0x0409: "Prepare queue full",
    0x040A: "Attribute not found",
    0x040B: "Attribute not long",
    0x040C: "Insufficient encryption key size",
    0x040D: "Invalid attribute value length",
    0x040E: "Unlikely error",
    0x040F: "Insufficient encryption",
    0x0410: "Unsupported group type",
    0x0411: "Insufficient resources",
    0x0480: "Application error codes",
}


def get_return_message(return_code):
    try:
        return return_codes[return_code]
    except KeyError:
        return "Unknown return code %04x" % return_code



"""
Exceptions for pygatt Module.
"""


class BLEError(Exception):
    """Exception class for pygatt."""
    pass


class NotConnectedError(BLEError):
    pass


class NotificationTimeout(BLEError):
    def __init__(self, msg=None, gatttool_output=None):
        super(NotificationTimeout, self).__init__(msg)
        self.gatttool_output = gatttool_output



from __future__ import print_function

import threading
import logging
from collections import defaultdict
from binascii import hexlify
from uuid import UUID

from . import exceptions

try:
    string_type = basestring
except NameError:
    string_type = str

log = logging.getLogger(__name__)


class BLEDevice(object):
    """
    An BLE device connection instance, returned by one of the BLEBackend
    implementations. This class is not meant to be instantiated directly - use
    BLEBackend.connect() to create one.
    """
    def __init__(self, address):
        """
        Initialize.

        address -- the BLE address (aka MAC address) of the device as a string.
        """
        self._address = address
        self._characteristics = {}
        self._callbacks = defaultdict(set)
        self._subscribed_handlers = {}
        self._subscribed_uuids = {}
        self._lock = threading.Lock()

    def bond(self, permanent=False):
        """
        Create a new bond or use an existing bond with the device and make the
        current connection bonded and encrypted.
        """
        raise NotImplementedError()

    def get_rssi(self):
        """
        Get the receiver signal strength indicator (RSSI) value from the BLE
        device.

        Returns the RSSI value in dBm on success.
        Returns None on failure.
        """
        raise NotImplementedError()

    def char_read(self, uuid):
        """
        Reads a Characteristic by UUID.

        uuid -- UUID of Characteristic to read as a string.

        Returns a bytearray containing the characteristic value on success.

        Example:
            my_ble_device.char_read('a1e8f5b1-696b-4e4c-87c6-69dfe0b0093b')
        """
        raise NotImplementedError()

    def char_read_handle(self, uuid):
        """
        Reads a Characteristic by handle.

        handle -- handle of Characteristic to read.

        Returns a bytearray containing the characteristic value on success.

        Example:
            my_ble_device.char_read_handle(5)
        """
        raise NotImplementedError()

    def char_read_long(self, uuid):
        """
        Reads a Characteristic by UUID.

        uuid -- UUID of Characteristic to read as a string.

        Returns a bytearray containing the characteristic value on success.

        Example:
            my_ble_device.char_read('a1e8f5b1-696b-4e4c-87c6-69dfe0b0093b')
        """
        raise NotImplementedError()

    def char_read_long_handle(self, handle):
        """
        Reads a Characteristic longer than one read by handle.

        handle -- handle of Characteristic to read.

        Returns a bytearray containing the characteristic value on success.

        Example:
            my_ble_device.char_read_long_handle(5)
        """
        raise NotImplementedError()

    def char_write(self, uuid, value, wait_for_response=True):
        """
        Writes a value to a given characteristic UUID.

        uuid -- the UUID of the characteristic to write to.
        value -- a bytearray to write to the characteristic.
        wait_for_response -- wait for response after writing. A GATT "command"
            is used when not waiting for a response. The remote host will not
            acknowledge the write.

        Example:
            my_ble_device.char_write('a1e8f5b1-696b-4e4c-87c6-69dfe0b0093b',
                                     bytearray([0x00, 0xFF]))
        """
        return self.char_write_handle(self.get_handle(uuid), value,
                                      wait_for_response=wait_for_response)

    def char_write_handle(self, handle, value, wait_for_response=True):
        """
        Writes a value to a given characteristic handle. This can be used to
        write to the characteristic config handle for a primary characteristic.

        hande -- the handle to write to.
        value -- a bytearray to write to the characteristic.
        wait_for_response -- wait for response after writing.

        Example:
            my_ble_device.char_write_handle(42, bytearray([0x00, 0xFF]))
        """
        raise NotImplementedError()

    def char_write_long(self, uuid, value, wait_for_response=False):
        """
        Writes a value to a given characteristic UUID.

        uuid -- the UUID of the characteristic to write to.
        value -- a bytearray to write to the characteristic.
        wait_for_response -- wait for response after writing.

        Example:
            my_ble_device.char_write('a1e8f5b1-696b-4e4c-87c6-69dfe0b0093b',
                                     bytearray([0x00, 0xFF]))
        """
        return self.char_write_long_handle(self.get_handle(uuid), value,
                                           wait_for_response=wait_for_response)

    def char_write_long_handle(self, handle, value, wait_for_response=False):
        """
        Writes a value to a given characteristic handle. This can be used to
        write to the characteristic config handle for a primary characteristic.

        hande -- the handle to write to.
        value -- a bytearray to write to the characteristic.
        wait_for_response -- wait for response after writing.

        Example:
            my_ble_device.char_write(42,
                                     bytearray([0x00, 0xFF]))
        """
        raise NotImplementedError()

    def disconnect(self):
        """
        Disconnect from the device. This instance of BLEDevice cannot be used
        after calling this method, you must call BLEBackend.connect() again to
        get a fresh connection.
        """
        raise NotImplementedError()

    def _notification_handles(self, uuid):
        # Expect notifications on the value handle...
        value_handle = self.get_handle(uuid)

        # but write to the characteristic config to enable notifications
        # TODO with the BGAPI backend we can be smarter and fetch the actual
        # characteristic config handle - we can also do that with gattool if we
        # use the 'desc' command, so we'll need to change the "get_handle" API
        # to be able to get the value or characteristic config handle.
        characteristic_config_handle = value_handle + 1

        return value_handle, characteristic_config_handle

    def subscribe(self, uuid, callback=None, indication=False):
        """
        Enable notifications or indications for a characteristic and register a
        callback function to be called whenever a new value arrives.

        uuid -- UUID as a string of the characteristic to subscribe.
        callback -- function to be called when a notification/indication is
                    received on this characteristic.
        indication -- use indications (where each notificaiton is ACKd). This is
                      more reliable, but slower.
        """

        value_handle, characteristic_config_handle = (
            self._notification_handles(uuid)
        )

        properties = bytearray([
            0x2 if indication else 0x1,
            0x0
        ])

        with self._lock:
            if callback is not None:
                self._callbacks[value_handle].add(callback)

            if self._subscribed_handlers.get(value_handle, None) != properties:
                self.char_write_handle(
                    characteristic_config_handle,
                    properties,
                    wait_for_response=False
                )
                log.info("Subscribed to uuid=%s", uuid)
                self._subscribed_handlers[value_handle] = properties
                self._subscribed_uuids[uuid] = indication
            else:
                log.debug("Already subscribed to uuid=%s", uuid)

    def unsubscribe(self, uuid):
        """
        Disable notification for a charecteristic and de-register the callback.
        """
        value_handle, characteristic_config_handle = (
            self._notification_handles(uuid)
        )

        properties = bytearray([0x0, 0x0])

        with self._lock:
            if uuid in self._subscribed_uuids:
                del(self._subscribed_uuids[uuid])
            if value_handle in self._callbacks:
                del(self._callbacks[value_handle])
            if value_handle in self._subscribed_handlers:
                del(self._subscribed_handlers[value_handle])
                self.char_write_handle(
                    characteristic_config_handle,
                    properties,
                    wait_for_response=False
                )
                log.info("Unsubscribed from uuid=%s", uuid)
            else:
                log.debug("Already unsubscribed from uuid=%s", uuid)

    def get_handle(self, char_uuid):
        """
        Look up and return the handle for an attribute by its UUID.
        :param char_uuid: The UUID of the characteristic.
        :type uuid: str
        :return: None if the UUID was not found.
        """
        if isinstance(char_uuid, string_type):
            char_uuid = UUID(char_uuid)
        log.debug("Looking up handle for characteristic %s", char_uuid)
        if char_uuid not in self._characteristics:
            self._characteristics = self.discover_characteristics()

        characteristic = self._characteristics.get(char_uuid)
        if characteristic is None:
            message = "No characteristic found matching %s" % char_uuid
            log.warn(message)
            raise exceptions.BLEError(message)

        # TODO support filtering by descriptor UUID, or maybe return the whole
        # Characteristic object
        log.debug("Found %s" % characteristic)
        return characteristic.handle

    def receive_notification(self, handle, value):
        """
        Receive a notification from the connected device and propagate the value
        to all registered callbacks.
        """

        log.info('Received notification on handle=0x%x, value=0x%s',
                 handle, hexlify(value))
        with self._lock:
            if handle in self._callbacks:
                for callback in self._callbacks[handle]:
                    callback(handle, value)

    def exchange_mtu(self, mtu):
        """
        ATT exchange Maximum Transmission Unit.
        :param mtu: New MTU-value
        :return: New MTU, as recognized by server.
        """
        raise NotImplementedError()

    def resubscribe_all(self):
        """
        Reenable all notifications and indications for uuids that were
        previously subscribed to.
        This has to be called after a connection loss and subsequent reconnect.
        """

        for uuid in self._subscribed_uuids:
            value_handle, characteristic_config_handle = (
                self._notification_handles(uuid)
            )

            properties = bytearray([
                0x2 if self._subscribed_uuids[uuid] else 0x1,
                0x0
            ])

            with self._lock:
                self.char_write_handle(
                    characteristic_config_handle,
                    properties,
                    wait_for_response=False
                )
                log.info("Resubscribed to uuid=%s", uuid)



from __future__ import print_function

import functools
import itertools
import re
import logging
import platform
import signal
import sys
import time
import threading
import subprocess
from uuid import UUID
from contextlib import contextmanager

from pygatt.exceptions import NotConnectedError, BLEError, NotificationTimeout
from pygatt.backends import BLEBackend, Characteristic, BLEAddressType
from pygatt.backends.backend import DEFAULT_CONNECT_TIMEOUT_S
from .device import GATTToolBLEDevice

DEFAULT_RECONNECT_DELAY = 1.0

log = logging.getLogger(__name__)


def _hex_value_parser(x):
    return bytearray.fromhex(x)


def is_windows():
    return platform.system() == 'Windows'

try:
    import pexpect
except Exception as err:
    if not is_windows():
        print("WARNING:", err, file=sys.stderr)


def at_most_one_device(func):
    """Every connection-specific function on the backend takes an instance of
    GATTToolBLEDevice as the first argument - this decorator will raise an
    exception if that device is not what the backend thinks is the currently
    connected device.
    """
    @functools.wraps(func)
    def wrapper(self, connected_device, *args, **kwargs):
        if connected_device != self._connected_device:
            raise NotConnectedError()
        return func(self, *args, **kwargs)
    return wrapper


class GATTToolReceiver(threading.Thread):
    """
    Observe pygatttool stdout in separate thread and dispatch events /
    callbacks.
    """

    def __init__(self, connection, parent_aliveness):
        super(GATTToolReceiver, self).__init__()
        self.daemon = True
        self._connection = connection
        self._parent_aliveness = parent_aliveness
        self._event_vector = {
            'notification': {
                'patterns': [r'Notification handle = .*? \r'],
            },
            'indication': {
                'patterns': [r'Indication   handle = .*? \r'],
            },
            'disconnected': {
                'patterns': [
                    r'.*Disconnected',
                    r'.*Invalid file descriptor',
                ]
            },
            'char_written': {
                'patterns': [
                    r'Characteristic value (was )?written successfully',
                ]
            },
            'value': {
                'patterns': [r'value: .*? \r']
            },
            'value/descriptor': {
                'patterns': [r'value/descriptor: .*? \r']
            },
            'discover': {
                'patterns': [
                    r'handle: 0x([a-fA-F0-9]{4}), '
                    'char properties: 0x[a-fA-F0-9]{2}, '
                    'char value handle: 0x([a-fA-F0-9]{4}), '
                    'uuid: ([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]'
                    '{4}-[0-9a-f]{12})\r\n',  # noqa
                ]
            },
            'connect': {
                'patterns': [r'Connection successful.*\[LE\]>']
            },
            'mtu': {
                'patterns': [
                    r'MTU was exchanged successfully: (\d+)'
                ]
            }
        }

        for event in self._event_vector.values():
            event["event"] = threading.Event()
            event["before"] = None
            event["after"] = None
            event["match"] = None
            event["callback"] = []

    def run(self):
        items = sorted(itertools.chain.from_iterable(
            [[(pattern, event)
              for pattern in event["patterns"]]
             for event in self._event_vector.values()])
        )
        patterns = [item[0] for item in items]
        events = [item[1] for item in items]

        log.info('Running...')
        while self._parent_aliveness.is_set():
            try:
                event_index = self._connection.expect(patterns, timeout=.5)
            except pexpect.TIMEOUT:
                continue
            except (NotConnectedError, pexpect.EOF):
                self._event_vector["disconnected"]["event"].set()
                break
            event = events[event_index]
            event["before"] = self._connection.before
            event["after"] = self._connection.after
            event["match"] = self._connection.match
            event["event"].set()
            for clb in event["callback"]:
                clb(event)
        log.info("Listener thread finished")

    def clear(self, event):
        """
        Clear event
        """
        self._event_vector[event]["event"].clear()

    def is_set(self, event):
        return self._event_vector[event]["event"].is_set()

    def wait(self, event, timeout=None):
        """
        Wait for event to be trigerred
        """
        if not self._event_vector[event]["event"].wait(timeout):
            raise NotificationTimeout()

    def register_callback(self, event, callback):
        """
        Call the callback function when event happens. Event wrapper
        is passed as argument.
        """
        self._event_vector[event]["callback"].append(callback)

    def remove_callback(self, event, callback):
        """
        Remove a registered callback, so it is no longer called when an
        event happens.
        """
        if callback in self._event_vector[event]["callback"]:
            self._event_vector[event]["callback"].remove(callback)

    def last_value(self, event, value_type):
        """
        Retrieve last value that saved by the event
        """
        return self._event_vector[event][value_type]

    @contextmanager
    def event(self, event, timeout=None):
        """
        Clear an event, execute context and then wait for event

        >>> with gtr.event("connect", 10):
        >>>     gtb.send(connect_command)

        """
        self.clear(event)
        yield
        self.wait(event, timeout)


class GATTToolBackend(BLEBackend):
    """
    Backend to pygatt that uses BlueZ's interactive gatttool CLI prompt.
    """

    def __init__(self, hci_device='hci0', gatttool_logfile=None,
                 cli_options=None, search_window_size=None, max_read=None):
        """
        Initialize.

        hci_device -- the hci_device to use with GATTTool.
        gatttool_logfile -- an optional filename to store raw gatttool
                input and output.
        search_window_size -- integer (optional); size in bytes of the
                search window that is used by `pexpect.expect`. This value
                should not exceed max_read
        max_read -- integer; number of bytes to read into gatt buffer at
                a time. Defaults to ~2000
        """

        if is_windows():
            raise BLEError("The GATTToolBackend requires BlueZ, "
                           "which is not available in Windows")

        self._hci_device = hci_device
        self._cli_options = cli_options
        self._connected_device = None
        self._gatttool_logfile = gatttool_logfile
        self._receiver = None
        self._con = None  # gatttool interactive session
        self._characteristics = {}
        self._running = threading.Event()
        self._address = None
        self._send_lock = threading.Lock()
        self._auto_reconnect = False
        self._reconnecting = False
        self._search_window_size = search_window_size
        self._scan = None
        self._max_read = max_read

    def sendline(self, command):
        """
        send a raw command to gatttool
        """
        with self._send_lock:
            self._con.sendline(command)

    def supports_unbonded(self):
        return False

    def start(self, reset_on_start=True, initialization_timeout=3):
        """
        Run gatttool to prepare for sending commands and monitoring the CLI tool
        output.

        :param bool reset_on_start: Perhaps due to a bug in gatttol or pygatt,
            but if the bluez backend isn't restarted, it can sometimes lock up
            the computer when trying to make a connection to HCI device.
        :param int initialization_timeout: Seconds to wait for the gatttool
            prompt. This should appear almost instantly, but on some HCI devices
            it may take longer to start up.
        """
        if self._con and self._running.is_set():
            self.stop()

        self._running.set()

        if reset_on_start:
            # Without restarting, sometimes when trying to bond with the
            # GATTTool backend, the entire computer will lock up.
            self.reset()

        # Start gatttool interactive session for device
        args = [
            'gatttool',
            self._cli_options,
            '-i',
            self._hci_device,
            '-I'
        ]
        gatttool_cmd = ' '.join([arg for arg in args if arg])
        log.debug('gatttool_cmd=%s', gatttool_cmd)
        if self._max_read:
            self._con = pexpect.spawn(
                gatttool_cmd, logfile=self._gatttool_logfile,
                searchwindowsize=self._search_window_size,
                maxread=self._max_read
            )
        else:
            self._con = pexpect.spawn(
                gatttool_cmd, logfile=self._gatttool_logfile,
                searchwindowsize=self._search_window_size,
            )

        # Wait for the interactive prompt
        self._con.expect(r'\[LE\]>', timeout=initialization_timeout)

        # Start the notification receiving thread
        self._receiver = GATTToolReceiver(self._con, self._running)
        self._receiver.daemon = True
        self._receiver.register_callback("disconnected", self._disconnect)
        for event in ["notification", "indication"]:
            self._receiver.register_callback(
                event,
                self._handle_notification_string
            )
        self._receiver.start()

    def stop(self):
        """
        Disconnects any connected device, stops the background receiving thread
        and closes the spawned gatttool process.
        disconnect.
        """
        self.disconnect(self._connected_device)
        if self._running.is_set():
            log.info('Stopping')
        self._running.clear()

        if self._con and self._con.isalive():
            while True:
                if not self._con.isalive():
                    break
                self.sendline('exit')
                time.sleep(0.1)
            self._con.close()
            self._con = None

    def scan(self, timeout=10, run_as_root=False):
        """
        By default, scanning with gatttool requires root privileges.
        If you don't want to require root, you must add a few
        'capabilities' to your system. If you have libcap installed, run this to
        enable normal users to perform LE scanning:
            setcap 'cap_net_raw,cap_net_admin+eip' `which hcitool`

        If you do use root, the hcitool subprocess becomes more difficult to
        terminate cleanly, and may leave your Bluetooth adapter in a bad state.
        """

        cmd = 'hcitool -i %s lescan' % self._hci_device
        if run_as_root:
            cmd = 'sudo %s' % cmd

        log.info("Starting BLE scan")
        self._scan = scan = pexpect.spawn(cmd)
        # "lescan" doesn't exit, so we're forcing a timeout here:
        try:
            scan.expect('foooooo', timeout=timeout)
        except pexpect.EOF:
            before_eof = scan.before.decode('utf-8')
            if "No such device" in before_eof:
                message = "No BLE adapter found"
            elif "Set scan parameters failed: Input/output error" in before_eof:
                message = ("BLE adapter requires reset after a scan as root"
                           "- call adapter.reset()")
            else:
                message = "Unexpected error when scanning: %s" % before_eof
            log.error(message)
            raise BLEError(message)
        except pexpect.TIMEOUT:
            devices = {}
            for line in scan.before.decode('utf-8').split('\r\n'):
                if 'sudo' in line:
                    raise BLEError("Enable passwordless sudo for 'hcitool' "
                                   "before scanning")
                match = re.match(
                    r'(([0-9A-Fa-f][0-9A-Fa-f]:?){6}) (\(?.+\)?)', line)

                if match is not None:
                    address = match.group(1)
                    name = match.group(3)
                    if name == "(unknown)":
                        name = None

                    if address in devices:
                        if (devices[address]['name'] is None) and (name is not
                                                                   None):
                            log.info("Discovered name of %s as %s",
                                     address, name)
                            devices[address]['name'] = name
                    else:
                        log.info("Discovered %s (%s)", address, name)
                        devices[address] = {
                            'address': address,
                            'name': name
                        }
            log.info("Found %d BLE devices", len(devices))
            return [device for device in devices.values()]
        finally:
            self.kill()
        return []

    def kill(self):
        if self._scan is None:
            return
        # Wait for lescan to exit cleanly, otherwise it leaves the BLE
        # adapter in a bad state and the device must be reset through BlueZ.
        # This will not work if run_as_root was used, since this process
        # itself doesn't have permission to terminate a process running as
        # root (hcitool itself). We recommend using the setcap tool to allow
        # scanning as a non-root user:
        #
        #    $ sudo setcap 'cap_net_raw,cap_net_admin+eip' `which hcitool`
        try:
            self._scan.kill(signal.SIGINT)
            self._scan.wait()
        except OSError:
            log.error("Unable to gracefully stop the scan - "
                      "BLE adapter may need to be reset.")

    def connect(self, address, timeout=DEFAULT_CONNECT_TIMEOUT_S,
                address_type=BLEAddressType.public, auto_reconnect=False):
        log.info('Connecting to %s with timeout=%s', address, timeout)
        self.sendline('sec-level low')
        self._address = address
        self._auto_reconnect = auto_reconnect

        try:
            cmd = 'connect {0} {1}'.format(self._address, address_type.name)
            with self._receiver.event("connect", timeout):
                self.sendline(cmd)
        except NotificationTimeout:
            message = "Timed out connecting to {0} after {1} seconds.".format(
                self._address, timeout
            )
            log.error(message)
            raise NotConnectedError(message)

        self._connected_device = GATTToolBLEDevice(address, self)
        return self._connected_device

    def clear_bond(self, address=None):
        """Use the 'bluetoothctl' program to erase a stored BLE bond.
        """
        con = pexpect.spawn('sudo bluetoothctl')

        try:
            con.expect("bluetooth", timeout=1)
            log.info("Clearing bond for %s", address)
            con.sendline("remove " + address.upper())
            con.expect(
                ["Device has been removed", "# "],
                timeout=.5
            )
        except pexpect.TIMEOUT:
            log.error("Unable to remove bonds for %s: %s",
                      address, con.before)
        finally:
            con.close(True)
        log.info("Removed bonds for %s", address)

    def _disconnect(self, event):
        if self._connected_device is not None and self._auto_reconnect:

            # this is called as a callback from the pexpect thread
            # the reconnection process has to be started in parallel, otherwise
            # the call is never finished
            log.info("Connection to %s lost. Reconnecting...", self._address)
            reconnect_thread = threading.Thread(target=self.reconnect,
                                                args=(self._connected_device, ))
            reconnect_thread.start()
        else:
            try:
                self.disconnect(self._connected_device)
            except NotConnectedError:
                pass

    @at_most_one_device
    def reconnect(self, timeout=DEFAULT_CONNECT_TIMEOUT_S):
        while self._auto_reconnect:
            log.info("Connecting to %s with timeout=%s", self._address,
                     timeout)
            try:
                cmd = "connect"
                with self._receiver.event("connect", timeout):
                    self.sendline(cmd)
                # reenable all notifications
                self._connected_device.resubscribe_all()
                log.info("Connection to %s reestablished.")
                break  # finished reconnecting
            except NotificationTimeout:
                message = ("Timed out connecting to {0} after {1} seconds. "
                           "Retrying in {2} seconds".format(
                                self._address, timeout,
                                DEFAULT_RECONNECT_DELAY))
                log.info(message)
                time.sleep(DEFAULT_RECONNECT_DELAY)

    @at_most_one_device
    def disconnect(self, *args, **kwargs):
        self._auto_reconnect = False  # disables any running reconnection
        if not self._receiver.is_set("disconnected"):
            self.sendline('disconnect')
        self._connected_device = None
        # TODO maybe call a disconnected callback on the device instance, so the
        # device knows if it was asynchronously disconnected?

    @at_most_one_device
    def bond(self, *args, **kwargs):
        log.info('Bonding')
        self.sendline('sec-level medium')

    def _save_charecteristic_callback(self, event):
        match = event["match"]
        try:
            value_handle = int(match.group(2), 16)
            char_uuid = match.group(3).strip().decode('ascii')
            self._characteristics[UUID(char_uuid)] = Characteristic(
                char_uuid, value_handle
            )
            log.debug(
                "Found characteristic %s, value handle: 0x%x",
                char_uuid,
                value_handle
            )
        except AttributeError:
            pass

    @at_most_one_device
    def discover_characteristics(self, timeout=5):
        self._characteristics = {}
        self._receiver.register_callback(
            "discover",
            self._save_charecteristic_callback,
        )
        self.sendline('characteristics')

        max_time = time.time() + timeout
        while not self._characteristics and time.time() < max_time:
            time.sleep(.5)

        # Sleep one extra second in case we caught characteristic
        # in the middle
        time.sleep(1)

        if not self._characteristics:
            raise NotConnectedError("Characteristic discovery failed")

        return self._characteristics

    def _handle_notification_string(self, event):
        msg = event["after"]
        if not msg:
            log.warn("Blank message received in notification, ignored")
            return

        match_obj = re.match(r'Notification handle = (0x[0-9a-f]+) value:(.*)',
                             msg.decode('utf-8'))
        if match_obj is None:
            log.warn("Unable to parse notification string, ignoring: %s", msg)
            return

        handle = int(match_obj.group(1), 16)
        values = _hex_value_parser(match_obj.group(2).strip())
        if self._connected_device is not None:
            self._connected_device.receive_notification(handle, values)

    @at_most_one_device
    def char_write_handle(self, handle, value, wait_for_response=True,
                          timeout=30):
        """
        Writes a value to a given characteristic handle.

        :param handle:
        :param value:
        :param wait_for_response: If true, performs an attribute write. If
            false, sends a command and expects no acknowledgement from the
            device.
        """
        cmd = 'char-write-{0} 0x{1:02x} {2}'.format(
            'req' if wait_for_response else 'cmd',
            handle,
            ''.join("{0:02x}".format(byte) for byte in value),
        )

        log.debug('Sending cmd=%s', cmd)
        if wait_for_response:
            try:
                with self._receiver.event("char_written", timeout=timeout):
                    self.sendline(cmd)
            except NotificationTimeout:
                log.error("No response received", exc_info=True)
                raise
        else:
            self.sendline(cmd)

        log.info('Sent cmd=%s', cmd)

    @at_most_one_device
    def char_read(self, uuid, timeout=1):
        """
        Reads a Characteristic by uuid.
        :param uuid: UUID of Characteristic to read.
        :type uuid: str
        :return: bytearray of result.
        :rtype: bytearray
        """
        with self._receiver.event("value", timeout=timeout):
            self.sendline('char-read-uuid %s' % uuid)
        rval = self._receiver.last_value("value", "after").split()[1:]
        return bytearray([int(x, 16) for x in rval])

    @at_most_one_device
    def char_read_handle(self, handle, timeout=4):
        """
        Reads a Characteristic by handle.
        :param handle: handle of Characteristic to read.
        :type handle: str
        :return: bytearray of result.
        :rtype: bytearray
        """
        with self._receiver.event("value/descriptor", timeout=timeout):
            self.sendline('char-read-hnd %s' % handle)
        rval = self._receiver.last_value("value/descriptor", "after"
                                         ).split()[1:]
        return bytearray([int(x, 16) for x in rval])

    @at_most_one_device
    def exchange_mtu(self, mtu, timeout=1):
        cmd = 'mtu {}'.format(mtu)

        log.debug('Requesting MTU: {}'.format(mtu))

        with self._receiver.event('mtu', timeout=timeout):
            self.sendline(cmd)
        try:
            rval = self._receiver.last_value("mtu", "after").split()[-1]
        except ValueError:
            log.error('MTU exchange failed: "{}"'.format(rval))
            raise

        log.debug('MTU exhange successful: {}'.format(rval))

        return rval

    def reset(self):
        subprocess.Popen(["sudo", "systemctl", "restart", "bluetooth"]).wait()
        subprocess.Popen([
            "sudo", "hciconfig", self._hci_device, "reset"]).wait()



import logging

from enum import Enum

log = logging.getLogger(__name__)

DEFAULT_CONNECT_TIMEOUT_S = 5.0

BLEAddressType = Enum('BLEAddressType', 'public random')


class BLEBackend(object):
    """Abstract base class representing a Bluetooth adapter backend. See the
    `pygatt.backends` module for available implementations.
    """

    def start(self):
        """Initialize and resource required to run the backend, e.g. background
        threads, USB device connections, etc.
        """
        raise NotImplementedError()

    def stop(self):
        """Stop and free any resources required while the backend is running.
        """
        raise NotImplementedError()

    def supports_unbonded(self):
        """Return True if the backend supports unbonded communication - this is
        to make detecting the GATTTool backend easier, which at the moment is
        auto-upgrading to a bonded connection even if not requested.
        """
        return True

    def connect(self, address, timeout=DEFAULT_CONNECT_TIMEOUT_S, **kwargs):
        """Return a BLEDevice for the connection if connected, otherwise raise
        an exception.
        """
        raise NotImplementedError()

    def scan(self, *args, **kwargs):
        """
        Performs a BLE scan.

        Returns a list of BLE devices found.
        """
        raise NotImplementedError()

    def filtered_scan(self, name_filter="", *args, **kwargs):
        """
        Scan for BLE devices and filter the list to include only with a name
        that includes the given filter.

        Returns a list of BLE devices found.
        """
        devices = self.scan(*args, **kwargs)
        return [device for device in devices
                if name_filter in (device['name'] or '')]

    def clear_bond(self, address=None):
        raise NotImplementedError()


class Characteristic(object):
    """
    A GATT characteristic, including it handle value and associated descriptors.
    Only valid for the lifespan of a BLE connection, since the handle values are
    dynamic.
    """
    def __init__(self, uuid, handle):
        """
        Sets the characteritic uuid and handle.

        handle - a bytearray
        """
        self.uuid = uuid
        self.handle = handle
        self.descriptors = {
            # uuid_string: handle
        }

    def add_descriptor(self, uuid, handle):
        """
        Add a characteristic descriptor to the dictionary of descriptors.
        """
        self.descriptors[uuid] = handle

    def __str__(self):
        return "<%s uuid=%s handle=%d>" % (self.__class__.__name__,
                                           self.uuid, self.handle)



import re
import logging
import serial.tools.list_ports

log = logging.getLogger(__name__)


class USBSerialDeviceInfo(object):
    """
    Contains the information about a usb device in an easy to use way.
    """
    device_name = None  # Name of the usb device
    port_name = None    # Name of serial port to which the device is connected
    vendor_id = None    # USB vendor id
    product_id = None   # USB product id

    def __str__(self):
        if self.vendor_id is None:
            vendor_id = str(self.vendor_id)
        else:
            vendor_id = '0x{:04x}'.format(self.vendor_id)
        if self.product_id is None:
            product_id = str(self.product_id)
        else:
            product_id = '0x{:04x}'.format(self.product_id)
        return (self.device_name + ' VID=' + vendor_id + ' PID=' + product_id +
                ' on ' + self.port_name)


def extract_vid_pid(info_string):
    """
    Try different methods of extracting vendor and product IDs from a string.

    The output from serial.tools.list_ports.comports() varies so
    widely from machine to machine and OS to OS, the only option we
    have right now is to add to a list of patterns that we have seen so
    far and compare the output.

    info_string -- the string possibly containing vendor and product IDs.

    Returns a tuple of (vendor ID, product ID) if a device is found.
    If an ID isn't found, returns None.
    """

    DEVICE_STRING_PATTERNS = [
        # '...VID:PID=XXXX:XXXX...'
        re.compile('.*VID:PID=([0-9A-Fa-f]{0,4}):([0-9A-Fa-f]{0,4}).*'),

        # '...VID_XXXX...PID_XXXX...'
        re.compile('.*VID_([0-9A-Fa-f]{0,4}).*PID_([0-9A-Fa-f]{0,4}).*')
    ]

    for p in DEVICE_STRING_PATTERNS:
        match = p.match(info_string)
        if match:
            return int(match.group(1), 16), int(match.group(2), 16)
    return None


def find_usb_serial_devices(vendor_id=None, product_id=None):
    """
    Discovers USB serial device(s) connected to the machine matching the input
    arguments. If no arguments are given or both are None, returns all devices
    found.

    vendor_id -- the USB vendor id to match.
    product_id -- the USB product id to match.

    Returns a list of USBDeviceInfo objects matching the criteria given.
    """
    devices = []
    raw_devices = list(serial.tools.list_ports.comports())
    log.debug("Found %d serial USB devices", len(raw_devices))
    for device in raw_devices:
        log.debug("Checking serial USB device: %s", device)
        dev = USBSerialDeviceInfo()
        dev.port_name = device[0]
        dev.device_name = device[1]
        found_device = extract_vid_pid(device[2])
        if found_device is not None:
            dev.vendor_id, dev.product_id = found_device
            if vendor_id is None and product_id is None:
                devices.append(dev)
            elif dev.vendor_id == vendor_id and product_id is None:
                devices.append(dev)
            elif dev.product_id == product_id and vendor_id is None:
                devices.append(dev)
            elif dev.product_id == product_id and dev.vendor_id == vendor_id:
                devices.append(dev)
            log.debug("USB device: %s", dev)
    return devices



__title__ = 'pygatt'
__version__ = '4.0.0'
__license__ = 'Apache License, Version 2.0 and MIT License'
__copyright__ = 'Copyright 2015 Stratos Inc. and Orion Labs'

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup  # pylint: disable=F0401,E0611

with open('README.rst') as f:
    readme = f.read()
with open('CHANGELOG.rst') as f:
    changelog = f.read()

setup(
    name=__title__,
    version=__version__,
    description='Python Bluetooth LE (Low Energy) and GATT Library',
    author='Chris Peplin <github@rhubarbtech.com>',
    author_email='github@rhubarbtech.com',
    packages=find_packages(exclude=("tests", "tests.*")),
    package_data={'': ['LICENSE']},
    license="Apache 2.0 and MIT",
    long_description=readme + '\n\n' + changelog,
    url='https://github.com/peplin/pygatt',
    install_requires=[
        'pyserial',
        'enum-compat'
    ],
    setup_requires=[
        'coverage >= 3.7.1',
        'nose >= 1.3.7'
    ],
    extras_require={
        'GATTTOOL': ["pexpect"],
    },
    package_dir={'pygatt': 'pygatt'},
    zip_safe=False,
    include_package_data=True,
    classifiers=(
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
    )
)



import logging


class NullHandler(logging.Handler):
    def emit(self, record):
        pass

# Initialize a null handler for logging to avoid printing spurious "No handlers
# could be found for logger" messages.
logging.getLogger(__name__).addHandler(NullHandler())

from .exceptions import BLEError  # noqa
from .device import BLEDevice  # noqa
from .backends import BGAPIBackend, GATTToolBackend, BLEAddressType  # noqa
